import os

from PySide6.QtGui import QIcon
from PySide6.QtWidgets import QLineEdit, QListWidgetItem

from OSCR import (
    compose_logfile, repair_logfile as oscr_repair_logfile, extract_bytes)

from .dialogs import confirmation_dialog, show_message
from .iofunctions import browse_path, save_to_json
from .textedit import format_path
from .translation import tr


def browse_log(self, entry: QLineEdit):
    """
    Callback for browse button.

    Parameters:
    - :param entry: QLineEdit -> path entry line widget
    """
    current_path = entry.text()
    if current_path != '':
        current_path = os.path.dirname(current_path)
    path = self.browse_path(current_path, 'Logfile (*.log);;Any File (*.*)')
    if path != '':
        entry.setText(format_path(path))
        if self.settings.auto_scan:
            self.analyze_log_callback(path=path, parser_num=1)


def save_combat(self, combat_info: tuple | None):
    """
    Callback for save button.

    Parameters:
    - :param combat_info: tuple of combat-identifying data (id, map, date, time, difficulty)
    """
    if combat_info is None or combat_info[0] >= len(self.parser.combats):
        return
    combat = self.parser.combats[combat_info[0]]
    if not combat:
        return
    filename = combat.map
    if combat.difficulty is not None and combat.difficulty != '':
        filename += ' ' + combat.difficulty
    filename += f' {combat.start_time.strftime("%Y-%m-%d %H.%M")}.log'
    base_dir = f'{os.path.dirname(self.entry.text())}/{filename}'
    path = browse_path(self, base_dir, 'Logfile (*.log);;Any File (*.*)', save=True)
    if path:
        self.parser.export_combat(combat_info[0], path)


def switch_analysis_tab(self, tab_index: int):
    """
    Callback for tab switch buttons; switches tab and sets active button.

    Parameters:
    - :param tab_index: index of the tab to switch to
    """
    self.widgets.analysis_graph_tabber.setCurrentIndex(tab_index)
    self.widgets.analysis_tree_tabber.setCurrentIndex(tab_index)
    for index, button in enumerate(self.widgets.analysis_menu_buttons):
        if not index == tab_index:
            button.setChecked(False)
        else:
            button.setChecked(True)


def switch_overview_tab(self, tab_index: int):
    """
    Callback for tab switch buttons; switches tab and sets active button.

    Parameters:
    - :param tab_index: index of the tab to switch to
    """
    self.widgets.overview_tabber.setCurrentIndex(tab_index)
    for index, button in enumerate(self.widgets.overview_menu_buttons):
        if not index == tab_index:
            button.setChecked(False)
        else:
            button.setChecked(True)


def switch_main_tab(self, tab_index: int):
    """
    Callback for main tab switch buttons. Switches main and sidebar tabs.

    Parameters:
    - :param tab_index: index of the tab to switch to
    """
    SIDEBAR_TAB_CONVERSION = {
        0: 0,
        1: 0,
        2: 1,
        3: 2
    }
    self.widgets.main_tabber.setCurrentIndex(tab_index)
    self.widgets.sidebar_tabber.setCurrentIndex(SIDEBAR_TAB_CONVERSION[tab_index])
    if tab_index == 0:
        self.widgets.overview_table_button.show()
    else:
        self.widgets.overview_table_button.hide()
    if tab_index == 1:
        self.widgets.analysis_graph_button.show()
    else:
        self.widgets.analysis_graph_button.hide()


def add_favorite_ladder(self):
    """
    Adds a latter to favorites list. Updates settings
    """
    current_item = self.widgets.ladder_selector.currentItem()
    if current_item is not None:
        current_ladder_key = f'{current_item.text()}|{current_item.difficulty}'
        if current_ladder_key not in self.settings.favorite_ladders:
            self.settings.favorite_ladders.append(current_ladder_key)
            ladder_text, difficulty = current_ladder_key.split('|')
            if difficulty == 'None':
                difficulty = None
            item = QListWidgetItem(ladder_text)
            item.difficulty = difficulty
            if difficulty != 'Any' and difficulty is not None:
                icon = self.icons[f'TFO-{difficulty.lower()}']
                icon.addPixmap(icon.pixmap(18, 24), QIcon.Mode.Selected)
                item.setIcon(icon)
            self.widgets.favorite_ladder_selector.addItem(item)


def remove_favorite_ladder(self):
    """
    Adds a latter to favorites list. Updates settings
    """
    current_item = self.widgets.favorite_ladder_selector.currentItem()
    if current_item is not None:
        current_ladder_key = f'{current_item.text()}|{current_item.difficulty}'
        if current_ladder_key in self.settings.favorite_ladders:
            self.settings.favorite_ladders.remove(current_ladder_key)
            row = self.widgets.favorite_ladder_selector.row(current_item)
            self.widgets.favorite_ladder_selector.takeItem(row)


def set_graph_resolution_setting(self, new_value: int):
    """
    Calculates new_value / 10 and stores it to settings.

    Parameters:
    - :param new_value: data points per second
    """
    try:
        setting_value = round(new_value / 10, 1)
        self.settings.graph_resolution = setting_value
        return setting_value
    except (ValueError, ZeroDivisionError):
        return


def set_parser_opacity_setting(self, new_value: int) -> str:
    """
    Calculates new_value / 10 and stores it to settings.

    Parameters:
    - :param new_value: 20 times the opacity percentage
    """
    setting_value = round(new_value / 20, 2)
    self.settings.liveparser__window_opacity = setting_value
    return f'{setting_value:.2f}'


def set_ui_scale_setting(self, new_value: int) -> str:
    """
    Calculates new_value / 50 and stores it to settings.

    Parameters:
    - :param new_value: 50 times the ui scale percentage
    """
    setting_value = round(new_value / 50, 2)
    self.settings.ui_scale = setting_value
    return f'{setting_value:.2f}'


def set_live_scale_setting(self, new_value: int):
    """
    Calculates new_value / 50 and stores it to settings.

    Parameters:
    - :param new_value: 50 times the live scale percentage
    """
    setting_value = round(new_value / 50, 2)
    self.settings.liveparser__window_scale = setting_value
    return f'{setting_value:.2f}'


def set_sto_logpath_setting(self, entry: QLineEdit):
    """
    Formats and stores new logpath to settings.

    Parameters:
    - :param entry: the entry that holds the path
    """
    formatted_path = format_path(entry.text())
    self.settings.sto_log_path = formatted_path
    entry.setText(formatted_path)


def browse_sto_logpath(self, entry: QLineEdit):
    """
    Opens prompt to select new logpath and stores it to settings.

    Parameters:
    - :param entry: the entry that holds the path
    """
    current_path = entry.text()
    if not current_path:
        current_path = self.config.home_dir
    new_path = self.browse_path(os.path.dirname(current_path), 'Logfile (*.log);;Any File (*.*)')
    if new_path:
        formatted_path = format_path(new_path)
        self.settings.sto_log_path = formatted_path
        entry.setText(formatted_path)


def copy_live_data_callback(self):
    """
    Copies the data from the live parser table.
    """
    data_model = self.widgets.live_parser_table.model()
    cell_data = data_model._data
    output = list()
    name_index = 0 if self.settings.liveparser__player_display == 'Name' else 1
    if self.settings.liveparser__copy_kills:
        for row in cell_data:
            output.append(f"{row[0][name_index]}: {row[1]:,.2f} ({row[6]:.0f})")
        output = '{ OSCR } DPS (Kills): ' + ' | '.join(output)
    else:
        for row in cell_data:
            output.append(f"{row[0][name_index]}: {row[1]:,.2f}")
        output = '{ OSCR } DPS: ' + ' | '.join(output)
    self.app.clipboard().setText(output)


def expand_overview_table(self):
    """
    Shows the overview table
    """
    self.widgets.overview_table_frame.show()


def collapse_overview_table(self):
    """
    Hides the overview table
    """
    self.widgets.overview_table_frame.hide()


def expand_analysis_graph(self):
    """
    Shows the analysis graph
    """
    self.widgets.analysis_graph_tabber.show()
    self.settings.analysis_graph = True


def collapse_analysis_graph(self):
    """
    Hides the analysis graph
    """
    self.widgets.analysis_graph_tabber.hide()
    self.settings.analysis_graph = False


def confirm_trim_logfile(self):
    """
    Prompts the user to confirm whether the logfile should be trimmed
    """
    title = tr('Trim Logfile')
    text = tr(
            'Trimming the logfile will delete all combats except for the most recent combat. '
            'Continue?')
    if confirmation_dialog(self, title, text):
        success = trim_logfile(self)
        if success:
            show_message(self, title, tr('Logfile has been trimmed.'))
        else:
            show_message(self, title, tr('Trimming the logfile failed.'), 'error')


def trim_logfile(self) -> bool:
    """
    Removes all combats but the most recent one from a logfile

    :return: True if successful, False if not
    """
    log_path = os.path.abspath(self.entry.text())
    if self.parser.log_path == log_path:
        self.parser.export_combat(0, log_path)
    else:
        combats = self.parser.isolate_combats(log_path, 1)
        if len(combats) < 1:
            return False
        combat = combats[0]
        extract_bytes(log_path, log_path, combat[5], combat[6])
    return True


def repair_logfile(self):
    """
    Repairs current logfile.
    """
    log_path = os.path.abspath(self.entry.text())
    if os.path.isfile(log_path):
        res = oscr_repair_logfile(log_path, str(self.config.templog_folder_path))
        if res == '':
            show_message(self, tr('Repair Logfile'), tr('The Logfile has been repaired.'))
        elif res == 'PermissionError':
            error_message = ('The Logfile could not be repaired due to a permission error. '
                             'Make sure the selected logfile can be overwritten without elevated '
                             'permissions.')
            show_message(self, tr('Permission Error'), tr(error_message), 'error')
    else:
        show_message(
                self, tr('Repair Logfile'),
                tr('The Logfile you are trying to open does not exist.'), 'warning')


def extract_combats(self, selected_indices: list):
    """
    Extracts combats in `selected_indices` from current logfile and prompts the user to select a
    file to write them to.

    Parameters:
    - :param selected_indices: list of model indices refering to the selected combats
    """
    combat_intervals = list()
    for index in selected_indices:
        data = index.data()
        combat_intervals.append((data[5], data[6]))
    combat_intervals.sort(key=lambda element: element[0])
    source_path = self.entry.text()
    target_path = browse_path(
            self, os.path.dirname(source_path), 'Logfile (*.log);;Any File (*.*)', save=True)
    if target_path != '':
        compose_logfile(
                source_path, target_path, combat_intervals, str(self.config.templog_folder_path))
        show_message(self, tr('Split Logfile'), tr('Logfile has been saved.'))


def populate_split_combats_list(self, combat_list):
    """
    Isolates all combats in the current logfile and inserts them into `combat_list`

    Parameters:
    - :param combat_list: QListView with CombatModel to insert the isolated combats into
    """
    combats = self.parser.isolate_combats(self.entry.text())
    combat_list.model().set_items(combats)


def export_combat_json(self, combat_info: tuple | None):
    """
    Exports current combat to JSON file

    Parameters:
    - :param combat_info: tuple of combat-identifying data (id, map, date, time, difficulty)
    """
    if combat_info is None or combat_info[0] >= len(self.parser.combats):
        return
    combat = self.parser.combats[combat_info[0]]
    filename = combat.map
    if combat.difficulty is not None and combat.difficulty != '':
        filename += ' ' + combat.difficulty
    filename += f' {combat.start_time.strftime("%Y-%m-%d %H.%M")}.json'
    base_dir = f'{os.path.dirname(self.entry.text())}/{filename}'
    path = browse_path(self, base_dir, 'JSON File (*.json);;Any File (*.*)', save=True)
    if path:
        save_to_json(path, combat.get_export())



import os
from pathlib import Path

from PySide6.QtCore import QByteArray, QSettings


class OSCRConfig():
    def __init__(self):
        self.config_dir: Path = Path()
        self.default_icon_size: int = 24
        self.default_live_parser_scale: float = 1.0
        self.default_ui_scale: float = 1.0
        self.excluded_event_ids: list[str] = ['Autodesc.Combatevent.Falling']
        self.home_dir: Path = Path()
        self.icon_size: int = 24
        self.link_downloads: str = 'https://github.com/STOCD/OSCR-UI/releases'
        self.link_github: str = 'https://github.com/STOCD/OSCR-UI'
        self.link_stobuilds: str = 'https://discord.gg/stobuilds'
        self.link_stocd: str = 'https://github.com/STOCD'
        self.link_website: str = 'https://oscr.stobuilds.com'
        self.live_graph_fields: tuple[str] = ('DPS', 'Debuff', 'Attacks-in Share', 'HPS')
        self.live_parser_scale: float = 1.0
        self.minimum_window_width: int = 1280
        self.minimum_window_height: int = 720
        self.settings_file: str = 'OSCR_UI_settings.ini'
        self.templog_folder_name: str = '_temp'
        self.templog_folder_path: Path = Path()
        self.ui_scale: float = 1.0

    def __repr__(self):
        return f'<OSCR-Config ui_scale={self.ui_scale} icon_size={self.icon_size} ...>'


class OSCRSettings():

    __slots__ = ('_settings', 'analysis_graph', 'auto_scan', 'combat_min_lines',
                 'combats_to_parse', 'copy_format', 'dmg_columns', 'favorite_ladders',
                 'first_overview_tab', 'graph_resolution', 'heal_columns', 'language', 'log_path',
                 'overview_sort_column', 'overview_sort_order', 'seconds_between_combats',
                 'sto_log_path', 'ui_scale', 'state__analysis_splitter', 'state__geometry',
                 'state__live_geometry', 'state__live_splitter', 'state__overview_splitter',
                 'liveparser__auto_enabled', 'liveparser__columns', 'liveparser__copy_kills',
                 'liveparser__graph_active', 'liveparser__graph_field',
                 'liveparser__player_display', 'liveparser__window_scale',
                 'liveparser__window_opacity')

    def __init__(self, settings_file_path: Path):
        self.analysis_graph: bool = True
        self.auto_scan: bool = False
        self.combat_min_lines: int = 20
        self.combats_to_parse: int = 10
        self.copy_format: str = 'Compact'
        self.dmg_columns: list[bool] = [True] * 21
        self.favorite_ladders: list[str] = list()
        self.first_overview_tab: int = 0
        self.graph_resolution: float = 0.2
        self.heal_columns: list[bool] = [True] * 13
        self.language: str = 'en'
        self.log_path: str = ''
        self.overview_sort_column: int = 1
        self.overview_sort_order: str = 'Descending'
        self.seconds_between_combats: int = 45
        self.sto_log_path: str = ''
        self.ui_scale: float = 1.0

        self.state__analysis_splitter: QByteArray = QByteArray()
        self.state__geometry: QByteArray = QByteArray()
        self.state__live_geometry: QByteArray = QByteArray()
        self.state__live_splitter: QByteArray = QByteArray()
        self.state__overview_splitter: QByteArray = QByteArray()

        self.liveparser__auto_enabled: bool = False
        self.liveparser__columns: list[bool] = [True, False, True, False, False, False, False]
        self.liveparser__copy_kills: bool = False
        self.liveparser__graph_active: bool = False
        self.liveparser__graph_field: int = 0
        self.liveparser__player_display: str = 'Handle'
        self.liveparser__window_scale: float = 1.0
        self.liveparser__window_opacity: float = 0.85

        if os.name == 'nt':
            self._settings = QSettings(str(settings_file_path), QSettings.Format.IniFormat)
        else:
            self._settings = QSettings(str(settings_file_path), QSettings.Format.NativeFormat)

        self.load_settings()

    def load_settings(self):
        """
        Loads settings from settings file given in constructor into attributes.
        """
        for setting in self.__slots__:
            if setting.startswith('_'):
                continue
            setting_id = setting.replace('__', '/')
            if self._settings.contains(setting_id):
                item_type = type(getattr(self, setting))
                if item_type is list:
                    settings_item: list = getattr(self, setting)
                    if len(settings_item) > 0:
                        list_element_type = type(settings_item[0])
                    else:
                        list_element_type = str
                    item_list = self._settings.value(setting_id, type=list)
                    if list_element_type is bool:
                        items = [True if el == 'true' else False for el in item_list]
                        setattr(self, setting, items)
                    else:
                        setattr(self, setting, [list_element_type(el) for el in item_list])
                else:
                    setattr(self, setting, self._settings.value(setting_id, type=item_type))

    def store_settings(self):
        """
        Stores settings from attributes to settings file given in constructor.
        """
        for setting in self.__slots__:
            if not setting.startswith('_'):
                setting_id = setting.replace('__', '/')
                self._settings.setValue(setting_id, getattr(self, setting))

    def set(self, setting_name: str, value):
        """
        Sets setting `setting_name` to `value`. Only use when direct assignment cannot be used
        (e.g. inside a lambda function).
        """
        setattr(self, setting_name, value)



from types import FunctionType, BuiltinFunctionType, MethodType
from typing import Callable

from PySide6.QtCore import QSize, Qt
from PySide6.QtWidgets import (
    QAbstractItemView, QComboBox, QFrame, QHBoxLayout, QHeaderView, QLabel, QLineEdit, QPushButton,
    QSizePolicy, QSlider, QTableView, QTreeView, QVBoxLayout)

from .style import get_style, get_style_class, merge_style, theme_font

CALLABLE = (FunctionType, BuiltinFunctionType, MethodType)

SMINMIN = QSizePolicy(QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Minimum)
SMAXMAX = QSizePolicy(QSizePolicy.Policy.Maximum, QSizePolicy.Policy.Maximum)
SMAXMIN = QSizePolicy(QSizePolicy.Policy.Maximum, QSizePolicy.Policy.Minimum)
SMINMAX = QSizePolicy(QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Maximum)
SMIXMAX = QSizePolicy(QSizePolicy.Policy.MinimumExpanding, QSizePolicy.Policy.Maximum)
SMIXMIN = QSizePolicy(QSizePolicy.Policy.MinimumExpanding, QSizePolicy.Policy.Minimum)
SMAX = QSizePolicy.Policy.Maximum
SMIN = QSizePolicy.Policy.Minimum
SEXPAND = QSizePolicy.Policy.Expanding
SFIXED = QSizePolicy.Policy.Fixed

ATOP = Qt.AlignmentFlag.AlignTop
ABOTTOM = Qt.AlignmentFlag.AlignBottom
ARIGHT = Qt.AlignmentFlag.AlignRight
ALEFT = Qt.AlignmentFlag.AlignLeft
ACENTER = Qt.AlignmentFlag.AlignCenter
AVCENTER = Qt.AlignmentFlag.AlignVCenter
AHCENTER = Qt.AlignmentFlag.AlignHCenter

RFIXED = QHeaderView.ResizeMode.Fixed

OVERTICAL = Qt.Orientation.Vertical

SMPIXEL = QAbstractItemView.ScrollMode.ScrollPerPixel

SCROLLOFF = Qt.ScrollBarPolicy.ScrollBarAlwaysOff
SCROLLON = Qt.ScrollBarPolicy.ScrollBarAlwaysOn


def create_button(self, text: str, style: str = 'button', style_override={}, toggle=None):
    """
    Creates a button according to style with parent.

    Parameters:
    - :param text: text to be shown on the button
    - :param style: name of the style as in self.theme or style dict
    - :param style_override: style dict to override default style (optional)
    - :param toggle: True or False when button should be a toggle button, None when it should be a \
    normal button; the bool value indicates the default state of the button

    :return: configured QPushButton
    """
    button = QPushButton(text)
    button.setStyleSheet(get_style_class(self, 'QPushButton', style, style_override))
    if 'font' in style_override:
        button.setFont(theme_font(self, style, style_override['font']))
    else:
        button.setFont(theme_font(self, style))
    button.setCursor(Qt.CursorShape.PointingHandCursor)
    button.setSizePolicy(SMAXMAX)
    if isinstance(toggle, bool):
        button.setCheckable(True)
        button.setChecked(toggle)
    return button


def create_icon_button(
        self, icon, tooltip: str = '', style: str = 'icon_button', parent=None,
        style_override={}, icon_size: tuple = ()) -> QPushButton:
    """
    Creates a button showing an icon according to style with parent.

    Parameters:
    - :param icon: icon to be shown on the button
    - :param tooltip: text to show when the mouse pointer is on the button
    - :param style: name of the style as in self.theme or style dict
    - :param parent: parent of the button (optional)
    - :param style_override: style dict to override default style (optional)
    - :param icon_size: set icon size in case it should be different from the default

    :return: configured QPushButton
    """
    button = QPushButton('', parent)
    button.setIcon(icon)
    if tooltip:
        button.setToolTip(tooltip)
    button.setStyleSheet(get_style_class(self, 'QPushButton', style, style_override))
    if len(icon_size) != 2:
        icon_size = [self.config.icon_size] * 2
    button.setIconSize(QSize(*icon_size))
    button.setCursor(Qt.CursorShape.PointingHandCursor)
    button.setSizePolicy(SMAXMAX)
    return button


def create_frame(self, style: str = 'frame', style_override: dict = {}, size_policy=None) -> QFrame:
    """
    Creates a frame with default styling

    Parameters:
    - :param style: style dict to override default style (optional)
    - :param size_policy: size policy of the frame (optional)

    :return: configured QFrame
    """
    frame = QFrame()
    frame.setStyleSheet(get_style(self, style, style_override))
    frame.setSizePolicy(size_policy if isinstance(size_policy, QSizePolicy) else SMAXMAX)
    return frame


def create_label(self, text: str, style: str = 'label', style_override={}) -> QLabel:
    """
    Creates a label according to style with parent.

    Parameters:
    - :param text: text to be shown on the label
    - :param style: name of the style as in self.theme
    - :param style_override: style dict to override default style (optional)

    :return: configured QLabel
    """
    label = QLabel()
    label.setText(text)
    label.setStyleSheet(get_style(self, style, style_override))
    label.setSizePolicy(SMAXMAX)
    if 'font' in style_override:
        label.setFont(theme_font(self, style, style_override['font']))
    else:
        label.setFont(theme_font(self, style))
    return label


def create_button_series(
        self, buttons: dict, style, shape: str = 'row', seperator: str = '', ret=False):
    """
    Creates a row / column of buttons.

    Parameters:
    - :param buttons: dictionary containing button details
        - key "default" contains style override for all buttons (optional)
        - all other keys represent one button, key will be the text on the button; value for the
        key contains dict with details for the specific button (all optional)
            - "callback": callable that will be called on button click
            - "style": individual style override dict
            - "toggle": True or False when button should be a toggle button, None when it should
                be a normal button; the bool value indicates the default state of the button
            - "stretch": stretch value for the button
            - "align": alignment flag for button
    - :param style: key for self.theme -> default style
    - :param shape: row / column
    - :param seperator: string seperator displayed between buttons (optional)

    :return: populated QVBoxlayout / QHBoxlayout
    """
    if 'default' in buttons:
        defaults = merge_style(self, self.theme[style], buttons.pop('default'))
    else:
        defaults = self.theme[style]

    if shape == 'column':
        layout = QVBoxLayout()
    else:
        shape = 'row'
        layout = QHBoxLayout()

    layout.setContentsMargins(0, 0, 0, 0)
    layout.setSpacing(0)

    button_list = []

    if seperator != '':
        sep_style = {
                'color': defaults['color'], 'margin': 0, 'padding': 0, 'background': '#00000000'}

    for i, (name, detail) in enumerate(buttons.items()):
        if 'style' in detail:
            button_style = merge_style(self, defaults, detail['style'])
        else:
            button_style = defaults
        toggle_button = detail['toggle'] if 'toggle' in detail else None
        bt = create_button(self, name, style, button_style, toggle_button)
        if 'callback' in detail and isinstance(detail['callback'], CALLABLE):
            if toggle_button:
                bt.clicked[bool].connect(detail['callback'])
            else:
                bt.clicked.connect(detail['callback'])
        stretch = detail['stretch'] if 'stretch' in detail else 0
        if 'align' in detail:
            layout.addWidget(bt, stretch, detail['align'])
        else:
            layout.addWidget(bt, stretch)
        button_list.append(bt)
        if seperator != '' and i < (len(buttons) - 1):
            sep_label = create_label(self, seperator, 'label', sep_style)
            sep_label.setSizePolicy(SMAXMIN)
            layout.addWidget(sep_label, alignment=ACENTER)

    if ret:
        return layout, button_list
    else:
        return layout


def create_combo_box(self, style: str = 'combobox', style_override: dict = {}) -> QComboBox:
    """
    Creates a combobox with given style and returns it.

    Parameters:
    - :param style: key for self.theme -> default style
    - :param style_override: style dict to override default style

    :return: styled QCombobox
    """
    combo_box = QComboBox()
    combo_box.setStyleSheet(get_style_class(self, 'QComboBox', style, style_override))
    if 'font' in style_override:
        combo_box.setFont(theme_font(self, style, style_override['font']))
    else:
        combo_box.setFont(theme_font(self, style))
    combo_box.setSizePolicy(SMINMAX)
    combo_box.setCursor(Qt.CursorShape.PointingHandCursor)
    combo_box.view().setCursor(Qt.CursorShape.PointingHandCursor)
    return combo_box


def create_entry(
        self, default_value='', validator=None, style: str = 'entry',
        style_override: dict = {}, placeholder='') -> QLineEdit:
    """
    Creates an entry widget and styles it.

    Parameters:
    - :param default_value: default value for the entry
    - :param validator: validator to validate entered characters against
    - :param style: key for self.theme -> default style
    - :param style_override: style dict to override default style
    - :param placeholder: placeholder shown when entry is empty

    :return: styled QLineEdit
    """
    entry = QLineEdit(default_value)
    entry.setValidator(validator)
    entry.setPlaceholderText(placeholder)
    entry.setStyleSheet(get_style_class(self, 'QLineEdit', style, style_override))
    if 'font' in style_override:
        entry.setFont(theme_font(self, style, style_override['font']))
    else:
        entry.setFont(theme_font(self, style))
    entry.setCursor(Qt.CursorShape.IBeamCursor)
    entry.setSizePolicy(SMAXMAX)
    return entry


def create_annotated_slider(
        self, default_value: int = 1, min: int = 0, max: int = 3,
        style: str = 'slider', style_override_slider: dict = {}, style_override_label: dict = {},
        callback: Callable = lambda v: v) -> QHBoxLayout:
    """
    Creates Slider with label to display the current value.

    Parameters:
    - :param default_value: start value for the slider
    - :param min: lowest value of the slider
    - :param max: highest value of the slider
    - :param style: key for self.theme -> default style
    - :param style_override_slider: style dict to override default style
    - :param style_override_label: style dict to override default style
    - :param callback: callable to be attached to the valueChanged signal of the slider; will be \
    passed value the slider was moved to; must return value that the label should be set to

    :return: layout with slider
    """
    def label_updater(new_value):
        if isinstance(callback, CALLABLE):
            new_text = callback(new_value)
            slider_label.setText(str(new_text))

    layout = QHBoxLayout()
    layout.setContentsMargins(0, 0, 0, 3)
    layout.setSpacing(self.theme['defaults']['margin'])
    slider_label = create_label(
            self, '', style, style_override=style_override_label)
    layout.addWidget(slider_label, alignment=AVCENTER)
    slider = QSlider(Qt.Orientation.Horizontal)
    slider.setRange(min, max)
    slider.setSingleStep(1)
    slider.setPageStep(1)
    slider.setValue(default_value)
    slider.setTickPosition(QSlider.TickPosition.NoTicks)
    slider.setFocusPolicy(Qt.FocusPolicy.WheelFocus)
    slider.setSizePolicy(SMINMAX)
    slider.setStyleSheet(get_style_class(self, 'QSlider', style, style_override_slider))
    slider.setFixedHeight(22)
    slider.valueChanged.connect(label_updater)
    layout.addWidget(slider, stretch=1, alignment=AVCENTER)
    label_updater(default_value)
    return layout


def resize_tree_table(tree: QTreeView):
    """
    Resizes the columns of the given tree table to fit its contents

    Parameters:
    - :param tree: QTreeView -> tree to be resized
    """
    for col in range(tree.header().count()):
        width = max(tree.sizeHintForColumn(col), tree.header().sectionSizeHint(col)) + 5
        tree.header().resizeSection(col, width)


def create_analysis_table(self, widget) -> QTreeView:
    """
    Creates and returns a QTreeView, styled according to widget.

    Parameters:
    - :param parent: parent of the table
    - :param widget: style key for the table

    :return: configured QTreeView
    """
    table = QTreeView()
    table.setStyleSheet(get_style_class(self, 'QTreeView', widget))
    table.setSizePolicy(SMINMIN)
    table.setAlternatingRowColors(True)
    table.setHorizontalScrollMode(SMPIXEL)
    table.setVerticalScrollMode(SMPIXEL)
    table.setSortingEnabled(True)
    table.setEditTriggers(QAbstractItemView.EditTrigger.NoEditTriggers)
    table.setSelectionMode(QAbstractItemView.SelectionMode.ExtendedSelection)
    table.setSelectionBehavior(QAbstractItemView.SelectionBehavior.SelectItems)
    table.header().setStyleSheet(get_style_class(self, 'QHeaderView', 'tree_table_header'))
    table.header().setSectionResizeMode(RFIXED)
    table.header().setMinimumSectionSize(1)
    table.header().setSectionsClickable(True)
    table.header().setStretchLastSection(False)
    table.header().setSortIndicatorShown(False)
    table.expanded.connect(lambda: resize_tree_table(table))
    table.collapsed.connect(lambda: resize_tree_table(table))
    return table


def style_table(self, table: QTableView, style_override: dict = {}, single_row_selection=False):
    """
    Styles the given table.

    Parameters:
    - :param table: table to be styled
    - :param style_override: style override for table
    - :param single_row_selection: True when only one row should be selectable at once
    """
    table.setAlternatingRowColors(self.theme['s.c']['table_alternate'])
    table.setShowGrid(self.theme['s.c']['table_gridline'])
    table.setSortingEnabled(True)
    table.setStyleSheet(get_style_class(self, 'QTableView', 'table', style_override))
    table.setHorizontalScrollMode(SMPIXEL)
    table.setVerticalScrollMode(SMPIXEL)
    table.horizontalHeader().setStyleSheet(get_style_class(self, 'QHeaderView', 'table_header'))
    table.verticalHeader().setStyleSheet(get_style_class(self, 'QHeaderView', 'table_index'))
    table.verticalHeader().setMinimumHeight(1)
    table.verticalHeader().setDefaultSectionSize(1)
    table.resizeColumnsToContents()
    table.resizeRowsToContents()
    table.horizontalHeader().setSortIndicatorShown(False)
    table.horizontalHeader().setSectionResizeMode(RFIXED)
    table.verticalHeader().setSectionResizeMode(RFIXED)
    table.setSizePolicy(SMINMIN)
    if single_row_selection:
        table.setSelectionMode(QAbstractItemView.SelectionMode.SingleSelection)
        table.setSelectionBehavior(QAbstractItemView.SelectionBehavior.SelectRows)



"""Backend interface to the OSCR web server"""

import gzip
import json
import os
import tempfile

from OSCR_django_client.api import (
        CombatlogApi, LadderApi, LadderEntriesApi, VariantApi)
from OSCR_django_client.api_client import ApiClient
from OSCR_django_client.exceptions import ServiceException
from PySide6.QtGui import QIcon
from PySide6.QtWidgets import QListWidgetItem, QMessageBox

from .callbacks import switch_main_tab, switch_overview_tab
from .datafunctions import CustomThread, analyze_log_callback
from .datamodels import LeagueTableModel, SortingProxy
from .dialogs import show_message
from .style import theme_font
from .subwindows import uploadresult_dialog
from .textedit import format_datetime_str
from .translation import tr

LEAGUE_TABLE_HEADER = (
        'Name', 'Handle', 'DPS', 'Total Damage', 'Deaths', 'Combat Time', 'Date', 'Max One Hit',
        'Debuff', 'Highest Damage Ability')

OSCR_SERVER_BACKEND = "https://oscr.stobuilds.com/"
# OSCR_SERVER_BACKEND = "http://127.0.0.1:8000"


def establish_league_connection(self):
    """
    Connects to the league server if not already connected.

    Parameters:
    - :param fetch_ladder: fetches available maps and updates map selector if true
    """
    if self.league_api is None:
        self.league_api = OSCRClient()
        map_fetch_thread = CustomThread(
            self.window, lambda: fetch_and_insert_maps(self)
        )
        map_fetch_thread.start()


def fetch_and_insert_maps(self):
    """
    Retrieves maps from API and inserts them into the list.
    """
    # Only populate the table once.
    if self.widgets.variant_combo.count() > 0:
        return

    variants = self.league_api.variants(ordering="-start_date")
    for variant in variants.results:
        self.widgets.variant_combo.addItem(variant.name)
        if variant.name == 'Default':
            self.widgets.variant_combo.setCurrentText('Default')


def update_seasonal_records(self, new_season: str):
    """
    Update the default records widget

    Parameters:
    - :param new_season: Name of the season to be shown
    """
    ladders = self.league_api.ladders(variant=new_season)
    if ladders is not None:
        self.widgets.ladder_selector.clear()
        for ladder in ladders.results:
            solo = "[Solo]" if ladder.is_solo else ""
            key = f"{ladder.name} {solo}|{ladder.difficulty}"
            text = f"{ladder.name} {solo}"
            self.league_api.ladder_dict[key] = ladder
            item = QListWidgetItem(text)
            item.difficulty = ladder.difficulty
            if ladder.difficulty != 'Any' and ladder.difficulty is not None:
                icon = self.icons[f'TFO-{ladder.difficulty.lower()}']
                icon.addPixmap(icon.pixmap(18, 24), QIcon.Mode.Selected)
                item.setIcon(icon)
            self.widgets.ladder_selector.addItem(item)


def apply_league_table_filter(self, filter_text: str):
    """
    Sets filter to proxy model of league table

    Parameters:
    - :param filter_text: text to filter the table for
    """
    try:
        self.widgets.ladder_table.model().name_filter = filter_text
    except AttributeError:
        pass


def slot_ladder(self, selected_map_item: QListWidgetItem):
    """
    Fetches current ladder and puts it into the table.

    Parameters:
    - :param selected_map_item: item containing name and difficulty of clicked map
    """
    map_key = f'{selected_map_item.text()}|{selected_map_item.difficulty}'
    if map_key not in self.league_api.ladder_dict:
        return

    selected_ladder = self.league_api.ladder_dict[map_key]
    self.league_api.current_ladder_id = selected_ladder.id
    ladder_data = self.league_api.ladder_entries(selected_ladder.id)
    if ladder_data.count > 50:
        self.league_api.entire_ladder_loaded = False
    else:
        self.league_api.entire_ladder_loaded = True
    self.league_api.pages_loaded = 1
    table_index = list()
    table_data = list()
    logfile_ids = list()

    for entry in ladder_data.results:
        logfile_ids.append(entry.combatlog)
        row = entry.data
        table_index.append(entry.rank)
        table_data.append(
            (
                row["name"],
                row["handle"],
                row["DPS"],
                row["total_damage"],
                row["deaths"],
                row["combat_time"],
                format_datetime_str(entry.var_date),
                row["max_one_hit"],
                row["debuff"],
                row.get("build", "Unknown"),
            )
        )

    model = LeagueTableModel(
        table_data,
        tr(LEAGUE_TABLE_HEADER),
        table_index,
        theme_font(self, "table_header"),
        theme_font(self, "table"),
        combatlog_id_list=logfile_ids,
    )
    sorting_proxy = SortingProxy()
    sorting_proxy.setSourceModel(model)
    table = self.widgets.ladder_table
    table.setModel(sorting_proxy)
    table.resizeColumnsToContents()
    table.scrollToTop()


def extend_ladder(self):
    """
    Extends the ladder table by 50 newly fetched rows.
    """
    if self.league_api.entire_ladder_loaded:
        return
    if self.league_api.current_ladder_id is None:
        return

    ladder_data = self.league_api.ladder_entries(
        self.league_api.current_ladder_id, self.league_api.pages_loaded + 1
    )
    if ladder_data is not None:
        if len(ladder_data.results) < 50:
            self.league_api.entire_ladder_loaded = True
        self.league_api.pages_loaded += 1
        table_index = list()
        table_data = list()
        logfile_ids = list()
        for entry in ladder_data.results:
            logfile_ids.append(entry.combatlog)
            row = entry.data
            table_index.append(entry.rank)
            table_data.append(
                (
                    row["name"],
                    row["handle"],
                    row["DPS"],
                    row["total_damage"],
                    row["deaths"],
                    row["combat_time"],
                    format_datetime_str(entry.var_date),
                    row["max_one_hit"],
                    row["debuff"],
                    row.get("build", "Unknown"),
                )
            )
        self.widgets.ladder_table.model().sourceModel().extend_data(
            table_index, table_data, logfile_ids
        )


def download_and_view_combat(self):
    """
    Download a combat log and view its contents in the overview / analysis pages.
    """
    table = self.widgets.ladder_table
    table_model = table.model().sourceModel()
    selection = table.selectedIndexes()
    original_index = table.model().mapToSource(selection[0])
    row = original_index.row()
    log_id = table_model._combatlog_id_list[row]
    result = self.league_api.download(log_id)
    result = gzip.decompress(result)

    with tempfile.NamedTemporaryFile(
        mode="wb", dir=str(self.config.templog_folder_path), delete=False
    ) as file:
        file.write(result)
    analyze_log_callback(
        self, path=file.name, hidden_path=True
    )
    switch_overview_tab(self, self.settings.first_overview_tab)
    switch_main_tab(self, 1)


def upload_callback(self):
    """
    Helper function to grab the current combat and upload it to the backend.
    """
    try:
        current_combat = self.parser.combats[self.current_combats.currentIndex().data()[0]]
    except IndexError:
        show_message(self, tr("Logfile Upload"), tr("No data to upload."), 'info')
        return

    establish_league_connection(self)

    with tempfile.NamedTemporaryFile(
            delete=False, dir=str(self.config.templog_folder_path)) as temp:
        with open(current_combat.log_file, 'rb') as log_file:
            log_file.seek(current_combat.file_pos[0])
            data = gzip.compress(
                    log_file.read(current_combat.file_pos[1] - current_combat.file_pos[0]))
            temp.write(data)
            temp.flush()
    res = self.league_api.upload(temp.name)
    if res:
        uploadresult_dialog(self, res)
    os.remove(temp.name)


class OSCRClient:
    def __init__(self):
        """Initialize an instance of the OSCR backlend client"""

        self.address = OSCR_SERVER_BACKEND
        self.api_client = ApiClient()
        self.api_client.configuration.host = self.address
        self.api_combatlog = CombatlogApi(api_client=self.api_client)
        self.api_ladder = LadderApi(api_client=self.api_client)
        self.api_ladder_entries = LadderEntriesApi(api_client=self.api_client)
        self.api_variant = VariantApi(api_client=self.api_client)
        self.ladder_dict: dict = dict()
        self.ladder_dict_season: dict = dict()
        self.current_ladder_id = None
        self.pages_loaded: int = -1
        self.entire_ladder_loaded: bool = True

    def upload(self, filename):
        """Upload a combat log located at path for analysis"""

        try:
            return self.api_combatlog.combatlog_uploadv2(file=filename)
        except ServiceException as e:
            reply = QMessageBox()
            reply.setWindowTitle("Open Source Combatlog Reader")
            try:
                data = json.loads(e.body)
                reply.setText(
                    data.get("detail", tr("Failed to parse error from server"))
                )
            except Exception:
                reply.setText(tr("Failed to parse error from server"))
            reply.exec()

    def download(self, id):
        """Download a combat log"""
        try:
            return self.api_combatlog.combatlog_download(id=id)
        except ServiceException as e:
            reply = QMessageBox()
            reply.setWindowTitle("Open Source Combatlog Reader")
            try:
                data = json.loads(e.body)
                reply.setText(
                    data.get("detail", tr("Failed to parse error from server"))
                )
            except Exception:
                reply.setText(tr("Failed to parse error from server"))
            reply.exec()

        return None

    def ladders(self, **kwargs):
        """Fetch the list of ladders"""
        try:
            return self.api_ladder.ladder_list(**kwargs)
        except ServiceException as e:
            reply = QMessageBox()
            reply.setWindowTitle("Open Source Combatlog Reader")
            try:
                data = json.loads(e.body)
                reply.setText(
                    data.get("detail", tr("Failed to parse error from server"))
                )
            except Exception:
                reply.setText(tr("Failed to parse error from server"))
            reply.exec()

        return None

    def ladder_entries(self, id, page=1):
        """Fetch the nth page of ladder entries"""
        try:
            return self.api_ladder_entries.ladder_entries_list(
                ladder=str(id),
                page=page,
                ordering="-data__DPS",
                page_size=50,
            )
        except ServiceException as e:
            reply = QMessageBox()
            reply.setWindowTitle("Open Source Combatlog Reader")
            try:
                data = json.loads(e.body)
                reply.setText(
                    data.get("detail", tr("Failed to parse error from server"))
                )
            except Exception:
                reply.setText(tr("Failed to parse error from server"))
            reply.exec()

        return None

    def variants(self, **kwargs):
        """Return a list of Variants"""

        try:
            return self.api_variant.variant_list(**kwargs)
        except ServiceException as e:
            reply = QMessageBox()
            reply.setWindowTitle("Open Source Combatlog Reader")
            try:
                data = json.loads(e.body)
                reply.setText(
                    data.get("detail", tr("Failed to parse error from server"))
                )
            except Exception:
                reply.setText(tr("Failed to parse error from server"))
            reply.exec()

        return None



import os
from threading import Thread

from PySide6.QtCore import Qt, QThread, Signal, Slot

from OSCR import HEAL_TREE_HEADER, TREE_HEADER
from OSCR.combat import Combat

from .callbacks import switch_main_tab, switch_overview_tab
from .datamodels import DamageTreeModel, HealTreeModel, TreeSelectionModel
from .dialogs import show_message
from .displayer import create_overview
from .textedit import format_damage_number, format_damage_tree_data, format_heal_tree_data
from .translation import tr


class CustomThread(QThread):
    """
    Subclass of QThread able to execute an arbitrary function in a seperate thread.
    """
    result = Signal(tuple)

    def __init__(self, parent, func) -> None:
        self._func = func
        super().__init__(parent)

    def run(self):
        r = self._func()
        self.result.emit((r,))


@Slot()
def analyze_log_callback(self, path=None, hidden_path=False):
    """
    Callback of "Analyze" button.

    Parameters:
    - :param path: path to combat log file
    - :param hidden_path: True when settings should not be updated with log path
    """
    if path == '' or not os.path.isfile(path):
        show_message(
                self, tr('Invalid Logfile'),
                tr('The Logfile you are trying to open does not exist.'), 'warning')
        return

    if self.thread is not None and self.thread.is_alive():
        return

    if not hidden_path and path != self.settings.log_path:
        self.settings.log_path = path

    self.parser.reset_parser()
    self.current_combats.model().clear()
    self.parser.log_path = path
    self.thread = Thread(target=self.parser.analyze_log_file, kwargs={'max_combats': 1})
    self.thread.start()

    # reset tabber
    switch_main_tab(self, 0)
    switch_overview_tab(self, self.settings.first_overview_tab)


def analyze_log_background(self, amount: int):
    """
    Analyzes older combats from current combatlog in the background.

    Parameters:
    - :param amount: amount of combats to analyze
    """
    if self.parser.bytes_consumed > 0 and self.thread is not None and not self.thread.is_alive():
        self.thread = Thread(target=self.parser.analyze_log_file_mp, kwargs={'max_combats': amount})
        self.thread.start()


def copy_summary_compact(self, combat_time, current_combat):
    parts = ["OSCR (DPS)", current_combat.map]
    difficulty = current_combat.difficulty
    if difficulty != 'Any' and difficulty is not None:
        parts.append(difficulty)
    parts.append(combat_time)
    players = sorted(
        current_combat.players.values(),
        reverse=True,
        key=lambda player: player.DPS,
    )
    for player in players:
        parts.append(f'{player.handle} - {player.DPS:,.0f}')
    return " | ".join(parts)


def copy_summary_verbose(self, combat_time, current_combat):
    summary = f'{{ OSCR }} {current_combat.map}'
    difficulty = current_combat.difficulty
    if difficulty is not None and difficulty != 'Any':
        summary += f' ({difficulty}) - DPS / DMG [{combat_time}]: '
    else:
        summary += f' - DPS / DMG [{combat_time}]: '
    players = sorted(
        current_combat.players.values(),
        reverse=True,
        key=lambda player: player.DPS,
    )
    parts = list()
    for player in players:
        parts.append(
                f"`{player.handle}` {player.DPS:,.0f} / "
                + format_damage_number(player.total_damage))
    return summary + " | ".join(parts)


def copy_summary_csv(self, combat_time, current_combat):
    difficulty = current_combat.difficulty
    if difficulty is None:
        difficulty = 'Unknown'
    parts = ["OSCR", "DPS", current_combat.map, difficulty, combat_time]
    players = sorted(
        current_combat.players.values(),
        reverse=True,
        key=lambda player: player.DPS,
    )
    for player in players:
        parts.append(player.handle)
        parts.append(f'{player.DPS:.0f}')
    return ', '.join(parts)


def copy_summary_callback(self):
    """
    Callback to set the combat summary of the active combat to the user's clipboard

    Parameters:
    - :param parser_num: which parser to take the data from
    """
    if self.current_combat_id < 0:
        return
    current_combat: Combat = self.parser.combats[self.current_combat_id]

    duration = current_combat.duration.total_seconds()
    combat_time = f'{int(duration / 60):02}:{duration % 60:02.0f}'

    copy_dispatch = {
        'Compact': copy_summary_compact,
        'Verbose': copy_summary_verbose,
        'CSV': copy_summary_csv,
    }
    try:
        summary = copy_dispatch.get(
            self.settings.copy_format, 'Compact'
        )(self, combat_time, current_combat)
    except Exception as e:
        print(e)

    self.app.clipboard().setText(summary)


def insert_combat(self, combat: Combat):
    """
    Called by parser as soon as combat has been analyzed. Inserts combat into UI.

    Parameters:
    - :param combat: analyzed combat
    """
    difficulty = combat.difficulty if combat.difficulty is not None else ''
    dt = combat.start_time
    date = f'{dt.year}-{dt.month:02d}-{dt.day:02d}'
    time = f'{dt.hour:02d}:{dt.minute:02d}:{dt.second:02d}'
    self.current_combats.model().insert_item((combat.id, combat.map, date, time, difficulty))
    if combat.id == 0:
        self.current_combats.setCurrentIndex(self.current_combats.model().createIndex(0, 0, 0))
        create_overview(self, combat)
        populate_analysis(self, combat)
        self.current_combat_id = 0
        analyze_log_background(self, self.settings.combats_to_parse - 1)


def analysis_data_slot(self, index: int):
    """
    Shows analyzed combat

    Parameters:
    - :param index: index of the combat in the parsers combat list
    """
    combat = self.parser.combats[index]
    create_overview(self, combat)
    populate_analysis(self, combat)
    self.current_combat_id = combat.id


def populate_analysis(self, combat: Combat):
    """
    Populates the Analysis' treeview table.

    Parameters:
    - :param combat: combat containing the data to show
    """
    damage_out_item, damage_in_item, heal_out_item, heal_in_item = combat.root_items

    damage_out_table = self.widgets.analysis_table_dout
    damage_out_model = DamageTreeModel(
            damage_out_item, self.theme_font('tree_table_header'), self.theme_font('tree_table'),
            self.theme_font('', self.theme['tree_table']['::item']['font']), tr(TREE_HEADER))
    damage_out_table.setModel(damage_out_model)
    damage_out_root_index = damage_out_model.createIndex(0, 0, damage_out_model._root)
    damage_out_table.expand(damage_out_model.index(0, 0, damage_out_root_index))
    damage_out_table.sortByColumn(1, Qt.SortOrder.AscendingOrder)
    damage_out_table.setSelectionModel(TreeSelectionModel(damage_out_model))

    damage_in_table = self.widgets.analysis_table_dtaken
    damage_in_model = DamageTreeModel(
            damage_in_item, self.theme_font('tree_table_header'), self.theme_font('tree_table'),
            self.theme_font('', self.theme['tree_table']['::item']['font']), tr(TREE_HEADER))
    damage_in_table.setModel(damage_in_model)
    damage_in_root_index = damage_in_model.createIndex(0, 0, damage_in_model._root)
    damage_in_table.expand(damage_in_model.index(0, 0, damage_in_root_index))
    damage_in_table.sortByColumn(1, Qt.SortOrder.AscendingOrder)
    damage_in_table.setSelectionModel(TreeSelectionModel(damage_in_model))

    heal_out_table = self.widgets.analysis_table_hout
    heal_out_model = HealTreeModel(
            heal_out_item, self.theme_font('tree_table_header'), self.theme_font('tree_table'),
            self.theme_font('', self.theme['tree_table']['::item']['font']),
            tr(HEAL_TREE_HEADER))
    heal_out_table.setModel(heal_out_model)
    heal_out_root_index = damage_in_model.createIndex(0, 0, heal_out_model._root)
    heal_out_table.expand(heal_out_model.index(0, 0, heal_out_root_index))
    heal_out_table.sortByColumn(1, Qt.SortOrder.AscendingOrder)
    heal_out_table.setSelectionModel(TreeSelectionModel(heal_out_model))

    heal_in_table = self.widgets.analysis_table_hin
    heal_in_model = HealTreeModel(
            heal_in_item, self.theme_font('tree_table_header'), self.theme_font('tree_table'),
            self.theme_font('', self.theme['tree_table']['::item']['font']),
            tr(HEAL_TREE_HEADER))
    heal_in_table.setModel(heal_in_model)
    heal_in_root_index = damage_in_model.createIndex(0, 0, heal_in_model._root)
    heal_in_table.expand(heal_in_model.index(0, 0, heal_in_root_index))
    heal_in_table.sortByColumn(1, Qt.SortOrder.AscendingOrder)
    heal_in_table.setSelectionModel(TreeSelectionModel(heal_in_model))

    update_shown_columns_dmg(self)
    update_shown_columns_heal(self)


def update_shown_columns_dmg(self):
    """
    Hides / shows columns of the dmg analysis tables.
    """
    dout_table = self.widgets.analysis_table_dout
    dtaken_table = self.widgets.analysis_table_dtaken
    for i, state in enumerate(self.settings.dmg_columns):
        if state:
            dout_table.showColumn(i + 1)
            dtaken_table.showColumn(i + 1)
        else:
            dout_table.hideColumn(i + 1)
            dtaken_table.hideColumn(i + 1)


def update_shown_columns_heal(self):
    """
    Hides / shows columns of the heal analysis tables.
    """
    hout_table = self.widgets.analysis_table_hout
    hin_table = self.widgets.analysis_table_hin
    for i, state in enumerate(self.settings.heal_columns):
        if state:
            hout_table.showColumn(i + 1)
            hin_table.showColumn(i + 1)
        else:
            hout_table.hideColumn(i + 1)
            hin_table.hideColumn(i + 1)


def resize_tree_table(tree):
    """
    Resizes the columns of the given tree table to fit its contents.

    Parameters:
    - :param tree: QTreeView -> tree to be resized
    """
    for col in range(tree.header().count()):
        width = max(tree.sizeHintForColumn(col), tree.header().sectionSizeHint(col)) + 5
        tree.header().resizeSection(col, width)


def copy_analysis_table_callback(self):
    """
    Copies the current selection of analysis table as tab-delimited table
    """
    if self.widgets.main_tabber.currentIndex() != 1:
        return
    current_tab = self.widgets.analysis_tree_tabber.currentIndex()
    current_table = self.widgets.analysis_table[current_tab]
    selection: list = current_table.selectedIndexes()
    if selection:
        selection.sort(key=lambda index: (index.row(), index.column()))
        output = list()
        last_row = -1
        for cell_index in selection:
            col = cell_index.column()
            if cell_index.row() != last_row:
                output.append(list())
            output[-1].append(cell_index.internalPointer().get_data(col))
            last_row = cell_index.row()
        output_text = '\n'.join(map(lambda row: '\t'.join(map(str, row)), output))
        self.app.clipboard().setText(output_text)


def copy_analysis_callback(self):
    """
    Callback for copy button on analysis tab
    """
    current_tab = self.widgets.analysis_tree_tabber.currentIndex()
    current_table = self.widgets.analysis_table[current_tab]
    copy_mode = self.widgets.analysis_copy_combobox.currentText()
    if copy_mode == tr('Selection'):
        if current_tab <= 1:
            current_header = tr(HEAL_TREE_HEADER)
            format_function = format_damage_tree_data
        else:
            current_header = tr(HEAL_TREE_HEADER)
            format_function = format_heal_tree_data
        selection = current_table.selectedIndexes()
        if selection:
            selection_dict = dict()
            for selected_cell in selection:
                column = selected_cell.column()
                row_name = selected_cell.internalPointer().get_data(0)
                if row_name not in selection_dict:
                    selection_dict[row_name] = dict()
                if column != 0:
                    cell_data = selected_cell.internalPointer().get_data(column)
                    selection_dict[row_name][column] = cell_data
            output = ['{ OSCR }']
            for row_name, row_data in selection_dict.items():
                formatted_row = list()
                for col, value in row_data.items():
                    formatted_row.append(f'[{current_header[col]}] {format_function(value, col)}')
                formatted_row_name = ''.join(row_name) if isinstance(row_name, tuple) else row_name
                output.append(f"`{formatted_row_name}`: {' | '.join(formatted_row)}")
            output_string = '\n'.join(output)
            self.app.clipboard().setText(output_string)
    elif copy_mode == tr('Global Max One Hit'):
        if current_tab <= 1:
            max_one_hit_col = 4
            prefix = tr('Max One Hit')
        else:
            max_one_hit_col = 7
            prefix = tr('Max One Heal')
        max_one_hits = []
        for player_item in current_table.model()._player._children:
            max_one_hits.append((player_item.get_data(max_one_hit_col), player_item))
        max_one_hit, max_one_hit_item = max(max_one_hits, key=lambda x: x[0])
        max_one_hit_ability = max(
                max_one_hit_item._children, key=lambda x: x.get_data(max_one_hit_col))
        max_one_hit_ability = max_one_hit_ability.get_data(0)
        if isinstance(max_one_hit_ability, tuple):
            max_one_hit_ability = ''.join(max_one_hit_ability)
        output_string = (f'{{ OSCR }} {prefix}: {max_one_hit:,.2f} '
                         f'(`{"".join(max_one_hit_item.get_data(0))}` – '
                         f'{max_one_hit_ability})')
        self.app.clipboard().setText(output_string)
    elif copy_mode == tr('Max One Hit'):
        if current_tab <= 1:
            max_one_hit_col = 4
            prefix = tr('Max One Hit')
        else:
            max_one_hit_col = 7
            prefix = tr('Max One Heal')
        selection = current_table.selectedIndexes()
        if selection:
            selected_row = selection[0].internalPointer()
            if selected_row._children:
                max_one_hit_item = max(
                        selected_row._children, key=lambda child: child.get_data(max_one_hit_col))
                max_one_hit = max_one_hit_item.get_data(max_one_hit_col)
                max_one_hit_ability = max_one_hit_item.get_data(0)
                if isinstance(max_one_hit_ability, tuple):
                    max_one_hit_ability = ''.join(max_one_hit_ability)
                output_string = (f'{{ OSCR }} {prefix}: {max_one_hit:,.2f} '
                                 f'(`{"".join(selected_row.get_data(0)[0:2])}` – '
                                 f'{max_one_hit_ability})')
                self.app.clipboard().setText(output_string)
    elif copy_mode == tr('Magnitude'):
        if current_tab == 0:
            prefix = tr('Total Damage Out')
        elif current_tab == 1:
            prefix = tr('Total Damage Taken')
        elif current_tab == 2:
            prefix = tr('Total Heal Out')
        else:
            prefix = tr('Total Heal In')
        magnitudes = list()
        for player_item in current_table.model()._player._children:
            magnitudes.append((player_item.get_data(2), ''.join(player_item.get_data(0)[:2])))
        magnitudes.sort(key=lambda x: x[0], reverse=True)
        magnitudes = [f"`[{player}]` {magnitude:,.2f}" for magnitude, player in magnitudes]
        output_string = (f'{{ OSCR }} {prefix}: {" | ".join(magnitudes)}')
        self.app.clipboard().setText(output_string)
    elif copy_mode == tr('Magnitude / s'):
        if current_tab == 0:
            prefix = tr('Total DPS Out')
        elif current_tab == 1:
            prefix = tr('Total DPS Taken')
        elif current_tab == 2:
            prefix = tr('Total HPS Out')
        else:
            prefix = tr('Total HPS In')
        magnitudes = list()
        for player_item in current_table.model()._player._children:
            magnitudes.append((player_item.get_data(1), ''.join(player_item.get_data(0)[:2])))
        magnitudes.sort(key=lambda x: x[0], reverse=True)
        magnitudes = [f"`[{player}]` {magnitude:,.2f}" for magnitude, player in magnitudes]
        output_string = (f'{{ OSCR }} {prefix}: {" | ".join(magnitudes)}')
        self.app.clipboard().setText(output_string)



import os
from traceback import format_exception

from PySide6.QtCore import QPoint, QSize, Qt
from PySide6.QtGui import QMouseEvent, QTextOption
from PySide6.QtWidgets import (
        QDialog, QGridLayout, QHBoxLayout, QListView, QMessageBox,
        QSplitter, QTableView, QTextEdit, QVBoxLayout)

from OSCR import LiveParser, LIVE_TABLE_HEADER

from .callbacks import (
        confirm_trim_logfile, copy_live_data_callback, extract_combats, populate_split_combats_list,
        repair_logfile)
from .dialogs import show_message
from .displayer import create_live_graph, update_live_display, update_live_graph, update_live_table
from .datamodels import CombatModel, LiveParserTableModel
from .iofunctions import open_link
from .style import get_style, get_style_class, theme_font
from .textedit import format_path
from .translation import tr
from .widgetbuilder import (
        create_button, create_button_series, create_frame, create_icon_button, create_label,
        ABOTTOM, AHCENTER, ALEFT, ARIGHT, ATOP, AVCENTER, RFIXED,
        SMAXMAX, SMINMAX, SMINMIN, SMIXMIN)
from .widgets import CombatDelegate, FlipButton, LiveParserWindow, SizeGrip


def split_dialog(self):
    """
    Opens dialog to split the current logfile.
    """
    main_layout = QVBoxLayout()
    thick = self.theme['app']['frame_thickness']
    main_layout.setContentsMargins(thick, thick, thick, thick)
    content_frame = create_frame(self)
    main_layout.addWidget(content_frame)
    current_logpath = self.entry.text()
    content_layout = QVBoxLayout()
    content_layout.setContentsMargins(thick, thick, thick, thick)
    content_layout.setSpacing(thick)
    log_layout = QHBoxLayout()
    log_layout.setContentsMargins(0, 0, 0, 0)
    log_layout.setSpacing(thick)
    log_layout.setAlignment(ALEFT)
    current_log_heading = create_label(
            self, tr('Selected Logfile:'), 'label_light')
    log_layout.addWidget(current_log_heading)
    current_log_label = create_label(
            self, format_path(current_logpath), 'label_subhead', {'margin-bottom': 0})
    log_layout.addWidget(current_log_label)
    content_layout.addLayout(log_layout)
    seperator = create_frame(self, style='hr', size_policy=SMINMAX)
    seperator.setFixedHeight(self.theme['hr']['height'])
    content_layout.addWidget(seperator)
    trim_layout = QGridLayout()
    trim_layout.setContentsMargins(0, 0, 0, 0)
    trim_layout.setSpacing(thick)
    trim_layout.setColumnStretch(0, 1)
    trim_heading = create_label(self, tr('Trim Logfile:'), 'label_heading')
    trim_layout.addWidget(trim_heading, 0, 0, alignment=ALEFT)
    label_text = tr(
            'Removes all combats except for the most recent one from the selected logfile. '
            'All previous combats will be lost!')
    trim_text = create_label(self, label_text)
    trim_text.setSizePolicy(SMINMAX)
    trim_text.setWordWrap(True)
    trim_layout.addWidget(trim_text, 1, 0)
    trim_button = create_button(self, tr('Trim'))
    trim_button.clicked.connect(lambda: confirm_trim_logfile(self))
    trim_layout.addWidget(trim_button, 0, 1, alignment=ARIGHT | ABOTTOM)
    content_layout.addLayout(trim_layout)
    seperator = create_frame(self, style='hr', size_policy=SMINMAX)
    seperator.setFixedHeight(self.theme['hr']['height'])
    content_layout.addWidget(seperator)
    repair_layout = QGridLayout()
    repair_layout.setContentsMargins(0, 0, 0, 0)
    repair_layout.setSpacing(thick)
    repair_layout.setColumnStretch(0, 1)
    repair_log_heading = create_label(self, tr('Repair Logfile:'), 'label_heading')
    repair_layout.addWidget(repair_log_heading, 0, 0, alignment=ALEFT)
    label_text = tr('Attempts to repair the logfile by replacing sections known to break parsing.')
    repair_label = create_label(self, label_text)
    repair_layout.addWidget(repair_label, 1, 0)
    repair_log_button = create_button(self, tr('Repair'))
    repair_log_button.clicked.connect(lambda: repair_logfile(self))
    repair_layout.addWidget(repair_log_button, 0, 1, alignment=ARIGHT | ABOTTOM)
    content_layout.addLayout(repair_layout)
    seperator = create_frame(self, style='hr', size_policy=SMINMAX)
    seperator.setFixedHeight(self.theme['hr']['height'])
    content_layout.addWidget(seperator)

    combat_list = QListView()
    split_heading_layout = QHBoxLayout()
    split_heading_layout.setContentsMargins(0, 0, 0, 0)
    split_heading_layout.setSpacing(thick)
    split_heading = create_label(self, tr('Split Logfile:'), 'label_heading')
    split_heading_layout.addWidget(split_heading, alignment=ALEFT, stretch=1)
    split_button_style = {
        tr('Load Combats'): {'callback': lambda: populate_split_combats_list(self, combat_list)},
        tr('Split'): {'callback': lambda: extract_combats(
                self, combat_list.selectionModel().selectedIndexes())},
    }
    buttons_layout = create_button_series(self, split_button_style, 'button', seperator='•')
    split_heading_layout.addLayout(buttons_layout)
    content_layout.addLayout(split_heading_layout)
    label_text = tr('Extracts (multiple) combats from selected file and saves them to new file.')
    split_label = create_label(self, label_text)
    content_layout.addWidget(split_label)
    background_frame = create_frame(self, style='frame', style_override={
            'border-radius': self.theme['listbox']['border-radius'], 'margin-top': '@csp',
            'margin-bottom': '@csp'}, size_policy=SMINMIN)
    background_layout = QVBoxLayout()
    background_layout.setContentsMargins(0, 0, 0, 0)
    background_frame.setLayout(background_layout)
    combat_list.setEditTriggers(QListView.EditTrigger.NoEditTriggers)
    combat_list.setSelectionMode(QListView.SelectionMode.MultiSelection)
    combat_list.setStyleSheet(get_style_class(self, 'QListView', 'listbox'))
    combat_list.setFont(theme_font(self, 'listbox'))
    combat_list.setAlternatingRowColors(True)
    combat_list.setSizePolicy(SMIXMIN)
    combat_list.setModel(CombatModel())
    ui_scale = self.config.ui_scale
    border_width = 1 * ui_scale
    padding = 4 * ui_scale
    combat_list.setItemDelegate(CombatDelegate(border_width, padding))
    background_layout.addWidget(combat_list)
    content_layout.addWidget(background_frame, alignment=AHCENTER)

    content_frame.setLayout(content_layout)

    dialog = QDialog(self.window)
    dialog.setLayout(main_layout)
    dialog.setWindowTitle(tr('OSCR - Split Logfile'))
    dialog.setStyleSheet(get_style(self, 'dialog_window'))
    dialog.setSizePolicy(SMAXMAX)
    dialog.exec()


def uploadresult_dialog(self, result: dict):
    """
    Shows a dialog that informs about the result of the triggered upload.

    Paramters:
    - :param result: dict containing result
    """
    dialog = QDialog(self.window)
    main_layout = QVBoxLayout()
    thick = self.theme['app']['frame_thickness']
    main_layout.setContentsMargins(thick, thick, thick, thick)
    content_frame = create_frame(self)
    main_layout.addWidget(content_frame)
    content_layout = QGridLayout()
    content_layout.setContentsMargins(thick, thick, thick, thick)
    content_layout.setSpacing(0)
    margin = {'margin-bottom': self.theme['defaults']['isp']}
    title_label = create_label(self, f"{result.detail}", 'label_heading', style_override=margin)
    content_layout.addWidget(title_label, 0, 0, 1, 4, alignment=ALEFT)
    view_button = create_button(self, 'View Online', style_override=margin)
    view_button.clicked.connect(lambda: view_upload_result(self, result.combatlog))
    if result.results:
        content_layout.addWidget(view_button, 0, 0, 1, 4, alignment=ARIGHT)
    icon_size = QSize(self.config.icon_size / 1.5, self.config.icon_size / 1.5)
    row = 0
    if result.results:
        for row, line in enumerate(result.results, 1):
            if row % 2 == 1:
                table_style = {'background-color': '@mbg', 'padding': (5, 3, 3, 3), 'margin': 0}
                icon_table_style = {'background-color': '@mbg', 'padding': 3, 'margin': 0}
            else:
                table_style = {'background-color': '@bg', 'padding': (5, 3, 3, 3), 'margin': 0}
                icon_table_style = {'background-color': '@bg', 'padding': 3, 'margin': 0}
            if line.updated:
                icon = self.icons['check'].pixmap(icon_size)
            else:
                icon = self.icons['dash'].pixmap(icon_size)
            status_label = create_label(self, '', style_override=icon_table_style)
            status_label.setPixmap(icon)
            status_label.setSizePolicy(SMINMIN)
            content_layout.addWidget(status_label, row, 0)
            name_label = create_label(self, line.name, style_override=table_style)
            name_label.setSizePolicy(SMINMAX)
            content_layout.addWidget(name_label, row, 1)
            value_label = create_label(self, str(line.value), style_override=table_style)
            value_label.setSizePolicy(SMINMAX)
            value_label.setAlignment(ARIGHT)
            content_layout.addWidget(value_label, row, 2)
            detail_label = create_label(self, line.detail, style_override=table_style)
            detail_label.setSizePolicy(SMINMAX)
            content_layout.addWidget(detail_label, row, 3)
    top_margin = {'margin-top': self.theme['defaults']['isp']}
    close_button = create_button(self, 'Close', style_override=top_margin)
    close_button.clicked.connect(dialog.close)
    content_layout.addWidget(close_button, row + 1, 0, 1, 4, alignment=AHCENTER)
    content_frame.setLayout(content_layout)

    dialog.setLayout(main_layout)
    dialog.setWindowTitle(tr('OSCR - Upload Results'))
    dialog.setStyleSheet(get_style(self, 'dialog_window'))
    dialog.setSizePolicy(SMAXMAX)
    dialog.setFixedSize(dialog.sizeHint())
    dialog.exec()


def live_parser_toggle(self, activate: bool):
    """
    Activates / Deactivates LiveParser.

    Parameters:
    - :param activate: True when parser should be shown; False when open parser should be closed.
    """
    if activate:
        log_path = self.settings.sto_log_path
        if not log_path or not os.path.isfile(log_path):
            show_message(self, tr('Invalid Logfile'), tr(
                    'Make sure to set the STO Logfile setting in the settings tab to a valid '
                    'logfile before starting the live parser.'), 'warning')
            self.widgets.live_parser_button.setChecked(False)
            return
        FIELD_INDEX_CONVERSION = {0: 0, 1: 2, 2: 3, 3: 4}
        graph_active = self.settings.liveparser__graph_active
        data_buffer = list()
        data_field = FIELD_INDEX_CONVERSION[self.settings.liveparser__graph_field]
        self.live_parser = LiveParser(log_path, update_callback=lambda p, t: update_live_display(
                self, p, t, graph_active, data_buffer, data_field),
                settings=self.live_parser_settings)
        create_live_parser_window(self)
    else:
        try:
            self.live_parser_window.close()
        except AttributeError:
            pass
        try:
            self.live_parser.stop()
        except AttributeError:
            pass
        self.live_parser_window.update_table.disconnect()
        self.live_parser_window.update_graph.disconnect()
        self.live_parser_window.deleteLater()
        self.live_parser_window = None
        self.live_parser = None
        self.widgets.live_parser_table = None
        self.widgets.live_parser_splitter = None
        self.widgets.live_parser_button.setChecked(False)


def create_live_parser_window(self):
    """
    Creates the LiveParser window.
    """
    self.config.live_parser_scale = self.settings.liveparser__window_scale
    ui_scale = self.config.ui_scale
    self.config.ui_scale = self.config.live_parser_scale

    live_window = LiveParserWindow()
    live_window.setStyleSheet(get_style(self, 'live_parser'))
    live_window.setWindowTitle("Live Parser")
    live_window.setWindowIcon(self.icons['oscr'])
    live_window.setWindowFlags(
            live_window.windowFlags()
            | Qt.WindowType.WindowStaysOnTopHint
            | Qt.WindowType.WindowDoesNotAcceptFocus
            | Qt.WindowType.FramelessWindowHint)
    # live_window.setAttribute(Qt.WidgetAttribute.WA_ShowWithoutActivating, True)
    live_window.setWindowOpacity(self.settings.liveparser__window_opacity)
    if self.settings.state__live_geometry:
        live_window.restoreGeometry(self.settings.state__live_geometry)
    live_window.closeEvent = lambda close_event: live_parser_close_callback(self, close_event)
    if self.app.platformName() == 'wayland':
        live_window.mousePressEvent = lambda event: live_parser_move_wayland(self, event)
    else:
        live_window.mousePressEvent = lambda press_event: live_parser_press_event(self, press_event)
        live_window.mouseMoveEvent = lambda move_event: live_parser_move_event(self, move_event)
    live_window.setSizePolicy(SMAXMAX)
    layout = QVBoxLayout()
    layout.setContentsMargins(0, 0, 0, 0)
    layout.setSpacing(0)

    graph_colors = None
    graph_column = None
    if self.settings.liveparser__graph_active:
        splitter = QSplitter(Qt.Orientation.Vertical)
        splitter.setStyleSheet(get_style_class(
                self, 'QSplitter', 'splitter', {'border': 'none', 'margin': 0}))
        splitter.setChildrenCollapsible(False)
        self.widgets.live_parser_splitter = splitter
        graph_frame, curves = create_live_graph(self)
        graph_frame.setMinimumHeight(self.sidebar_item_width * 0.1)
        splitter.addWidget(graph_frame)
        self.widgets.live_parser_curves = curves
        FIELD_INDEX_CONVERSION = {0: 0, 1: 2, 2: 3, 3: 4}
        graph_column = FIELD_INDEX_CONVERSION[self.settings.liveparser__graph_field]
        graph_colors = (*self.theme['plot']['color_cycler'][:5], '#eeeeee')
        layout.addWidget(splitter, stretch=1)

    table = QTableView()
    table.setAlternatingRowColors(self.theme['s.c']['table_alternate'])
    table.setShowGrid(self.theme['s.c']['table_gridline'])
    table.setStyleSheet(get_style_class(self, 'QTableView', 'live_table'))
    table.setVerticalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
    table.setHorizontalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
    table.horizontalHeader().setStyleSheet(
            get_style_class(self, 'QHeaderView', 'live_table_header'))
    table.verticalHeader().setStyleSheet(get_style_class(self, 'QHeaderView', 'live_table_index'))
    table.verticalHeader().setMinimumHeight(1)
    table.verticalHeader().setDefaultSectionSize(table.verticalHeader().fontMetrics().height() + 2)
    table.horizontalHeader().setMinimumWidth(1)
    table.horizontalHeader().setDefaultSectionSize(1)
    table.horizontalHeader().setSectionResizeMode(RFIXED)
    table.verticalHeader().setSectionResizeMode(RFIXED)
    table.setSizePolicy(SMINMIN)
    table.setSelectionMode(QTableView.SelectionMode.NoSelection)
    table.setMinimumWidth(self.sidebar_item_width * 0.1)
    table.setMinimumHeight(self.sidebar_item_width * 0.1)
    table.setSortingEnabled(True)
    if self.settings.liveparser__player_display == 'Handle':
        name_index = 1
    else:
        name_index = 0
    placeholder = [0] * len(LIVE_TABLE_HEADER)
    model = LiveParserTableModel(
            [[('Name', '@handle'), *placeholder, 0]], tr(LIVE_TABLE_HEADER), [],
            theme_font(self, 'live_table_header'), theme_font(self, 'live_table'),
            legend_col=graph_column, colors=graph_colors, name_index=name_index)
    table.setModel(model)
    table.resizeColumnsToContents()
    table.resizeRowsToContents()
    for index, state in enumerate(self.settings.liveparser__columns):
        if not state:
            table.hideColumn(index)
    self.widgets.live_parser_table = table
    if self.settings.liveparser__graph_active:
        splitter.addWidget(table)
        if self.settings.state__live_splitter:
            splitter.restoreState(self.settings.state__live_splitter)
    else:
        layout.addWidget(table, 1)

    margin = self.config.ui_scale * 6
    bottom_layout = QGridLayout()
    bottom_layout.setContentsMargins(margin, 0, 0, 0)
    bottom_layout.setSpacing(margin)
    bottom_layout.setColumnStretch(4, 1)

    activate_button = FlipButton(tr('Activate'), tr('Deactivate'), checkable=True)
    activate_button.setStyleSheet(self.get_style_class(
            'QPushButton', 'toggle_button', {'margin': (0, 0, 3, 0)}))
    activate_button.setFont(self.theme_font('app', '@subhead'))
    activate_button.r_function = lambda: self.live_parser.start()
    activate_button.l_function = lambda: self.live_parser.stop()
    bottom_layout.addWidget(activate_button, 0, 0, alignment=ALEFT | AVCENTER)
    icon_size = [self.theme['s.c']['button_icon_size'] * self.config.live_parser_scale * 0.8] * 2
    copy_button = create_icon_button(
            self, self.icons['copy'], tr('Copy Result'), style_override={'margin': (0, 0, 3, 0)},
            icon_size=icon_size)
    copy_button.clicked.connect(lambda: copy_live_data_callback(self))
    bottom_layout.addWidget(copy_button, 0, 1, alignment=ALEFT | AVCENTER)
    close_button = create_icon_button(
            self, self.icons['close'], tr('Close Live Parser'),
            style_override={'margin': (0, 0, 3, 0)}, icon_size=icon_size)
    close_button.clicked.connect(lambda: live_parser_toggle(self, False))
    bottom_layout.addWidget(close_button, 0, 2, alignment=ALEFT | AVCENTER)
    time_label = create_label(self, 'Duration: 0s')
    bottom_layout.addWidget(time_label, 0, 3, alignment=ALEFT | AVCENTER)
    self.widgets.live_parser_duration_label = time_label

    grip = SizeGrip(live_window)
    grip.setStyleSheet(get_style(self, 'resize_handle'))
    bottom_layout.addWidget(grip, 0, 4, alignment=ARIGHT | ABOTTOM)

    layout.addLayout(bottom_layout)
    live_window.setLayout(layout)
    live_window.update_table.connect(lambda data: update_live_table(self, data))
    live_window.update_graph.connect(update_live_graph)
    self.live_parser_window = live_window
    self.config.ui_scale = ui_scale
    live_window.show()

    if self.settings.liveparser__auto_enabled:
        activate_button.flip()


def live_parser_close_callback(self, event):
    """
    Executed when application is closed.
    """
    window_geometry = self.live_parser_window.saveGeometry()
    self.settings.state__live_geometry = window_geometry
    try:
        self.settings.state__live_splitter = self.widgets.live_parser_splitter.saveState()
    except AttributeError:
        pass
    event.accept()


def live_parser_press_event(self, event: QMouseEvent):
    """
    Used to start moving the parser window.
    """
    self.live_parser_window.start_pos = event.globalPosition().toPoint()
    event.accept()


def live_parser_move_event(self, event: QMouseEvent):
    """
    Used to move the parser window to new location.
    """
    parser_window = self.live_parser_window
    pos_delta = QPoint(event.globalPosition().toPoint() - parser_window.start_pos)
    parser_window.move(parser_window.x() + pos_delta.x(), parser_window.y() + pos_delta.y())
    parser_window.start_pos = event.globalPosition().toPoint()
    event.accept()


def live_parser_move_wayland(self, event: QMouseEvent):
    """
    Used to move the parser window on wayland.
    """
    self.live_parser_window.windowHandle().startSystemMove()
    event.accept()


def view_upload_result(self, log_id: str):
    """
    Opens webbrowser to show the uploaded combatlog on the DPS League tables.
    """
    open_link(f"https://oscr.stobuilds.com/ui/combatlog/{log_id}/")


def show_detection_info(self, combat_index: int):
    """
    Shows a subwindow containing information on the detection process

    Parameters:
    - :param combat_index: combat index in `self.parser.combats` identifying the combat to show \
    detection data on
    """
    if combat_index < 0:
        return
    dialog = QDialog(self.window)
    thick = self.theme['app']['frame_thickness']
    item_spacing = self.theme['defaults']['isp']
    main_layout = QVBoxLayout()
    main_layout.setContentsMargins(thick, thick, thick, thick)
    content_frame = create_frame(self)
    main_layout.addWidget(content_frame)
    content_layout = QVBoxLayout()
    content_layout.setContentsMargins(thick, thick, thick, thick)
    content_layout.setSpacing(item_spacing)

    for detection_info in self.parser.combats[combat_index].meta['detection_info']:
        if detection_info.success:
            if detection_info.step == 'existence':
                detection_method = tr('by checking whether the following entities exist in the log')
            elif detection_info.step == 'deaths':
                detection_method = tr('by checking the death counts of following entities')
            else:
                detection_method = tr('by checking the hull values of following entities')
            if detection_info.type == 'both':
                detected_type = tr('Map and Difficulty were')
            elif detection_info.type == 'difficulty':
                detected_type = f"{tr('Difficulty')} ({detection_info.difficulty}) {tr('was')}"
            else:
                detected_type = f"{tr('Map')} ({detection_info.map}) {tr('was')}"
            t = f"{tr('The')} {detected_type} {tr('successfully detected')} {detection_method}"
            t += ': ' + ', '.join(detection_info.identificators) + '.'
        else:
            if detection_info.type == 'both':
                detected_type = tr('Map and Difficulty')
            elif detection_info.type == 'difficulty':
                detected_type = f"{tr('Difficulty')} ({detection_info.difficulty})"
            else:
                detected_type = f"{tr('Map')} ({detection_info.map}) {tr('was')}"
            t = f"{tr('The')} {tr(detected_type)} {tr('could not be detected, because')} "
            if detection_info.step == 'existence':
                t += tr('no entity identifying a map was found in the log.')
            elif detection_info.step == 'deaths':
                t += f'{tr("the entity")} "{detection_info.identificators[0]}" {tr("was killed")} '
                t += f"{detection_info.retrieved_value} {tr('times instead of the expected')} "
                t += f"{detection_info.target_value} {tr('times')}."
            else:
                t += f'{tr("the entities")} "{detection_info.identificators[0]}" '
                t += f"{tr('average hull capacity of')} {detection_info.retrieved_value:.0f} "
                t += f"{tr('was higher than the allowed')} {detection_info.target_value:.0f}."
        info_label = create_label(self, t)
        info_label.setSizePolicy(SMINMAX)
        info_label.setWordWrap(True)
        content_layout.addWidget(info_label)

    seperator = create_frame(self, style='light_frame', size_policy=SMINMAX)
    seperator.setFixedHeight(1)
    content_layout.addWidget(seperator)
    ok_button = create_button(self, tr('OK'))
    ok_button.clicked.connect(lambda: dialog.done(0))
    content_layout.addWidget(ok_button, alignment=AHCENTER)
    content_frame.setLayout(content_layout)

    dialog = QDialog(self.window)
    dialog.setLayout(main_layout)
    dialog.setWindowTitle(tr('OSCR - Map Detection Details'))
    dialog.setStyleSheet(get_style(self, 'dialog_window'))
    dialog.setSizePolicy(SMAXMAX)
    dialog.exec()


def show_parser_error(self, error: BaseException):
    """
    Displays subwindow showing an error message and the given error traceback.

    - :param error: captured error with optionally additional data in the error.args attribute
    """
    default_message, *additional_messages = error.args
    error.args = (default_message,)
    error_text = ''.join(format_exception(error))
    if len(additional_messages) > 0:
        error_text += '\n\n++++++++++++++++++++++++++++++++++++++++++++++++++\n\n'
        error_text += '\n'.join(additional_messages)
    dialog = QDialog(self.window)
    thick = self.theme['app']['frame_thickness']
    item_spacing = self.theme['defaults']['isp']
    main_layout = QVBoxLayout()
    main_layout.setContentsMargins(thick, thick, thick, thick)
    dialog_frame = create_frame(self, size_policy=SMINMIN)
    main_layout.addWidget(dialog_frame)
    dialog_layout = QVBoxLayout()
    dialog_layout.setContentsMargins(thick, thick, thick, thick)
    dialog_layout.setSpacing(thick)
    content_frame = create_frame(self, size_policy=SMINMIN)
    content_layout = QVBoxLayout()
    content_layout.setContentsMargins(0, 0, 0, 0)
    content_layout.setSpacing(item_spacing)
    content_layout.setAlignment(ATOP)

    top_layout = QHBoxLayout()
    top_layout.setContentsMargins(0, 0, 0, 0)
    top_layout.setSpacing(2 * thick)
    icon_label = create_label(self, '')
    icon_size = self.theme['s.c']['big_icon_size'] * self.config.ui_scale
    icon_label.setPixmap(self.icons['error'].pixmap(icon_size))
    top_layout.addWidget(icon_label, alignment=ALEFT | AVCENTER)
    msg = tr(
            'An error occurred while parsing the selected combatlog. You can try repairing the '
            'log file using the repair functionality in the "Manage Logfile" dialog. If the error '
            'persists, please report it to the #oscr-support channel in the STOBuilds Discord.')
    message_label = create_label(self, msg)
    message_label.setWordWrap(True)
    message_label.setSizePolicy(SMINMAX)
    top_layout.addWidget(message_label, stretch=1)
    content_layout.addLayout(top_layout)
    error_field = QTextEdit()
    error_field.setSizePolicy(SMINMIN)
    error_field.setText(error_text)
    error_field.setReadOnly(True)
    error_field.setWordWrapMode(QTextOption.WrapMode.NoWrap)
    error_field.setFont(theme_font(self, 'textedit'))
    error_field.setStyleSheet(get_style_class(self, 'QTextEdit', 'textedit'))
    expand_button = FlipButton(tr('Show Error'), tr('Hide Error'))
    expand_button.set_icon_r(self.icons['chevron-right'])
    expand_button.set_icon_l(self.icons['chevron-down'])
    expand_button.r_function = error_field.show
    expand_button.l_function = error_field.hide
    expand_button.setStyleSheet(get_style_class(self, 'FlipButton', 'button'))
    expand_button.setFont(theme_font(self, 'button'))
    content_layout.addWidget(expand_button, alignment=ALEFT)
    content_layout.addWidget(error_field, stretch=1)
    error_field.hide()
    content_frame.setLayout(content_layout)
    dialog_layout.addWidget(content_frame, stretch=1)

    seperator = create_frame(self, style='light_frame', size_policy=SMINMAX)
    seperator.setFixedHeight(1)
    dialog_layout.addWidget(seperator)
    ok_button = create_button(self, tr('OK'))
    ok_button.clicked.connect(lambda: dialog.done(0))
    dialog_layout.addWidget(ok_button, alignment=AHCENTER)
    dialog_frame.setLayout(dialog_layout)

    dialog = QDialog(self.window)
    dialog.setLayout(main_layout)
    dialog.setWindowTitle(tr('OSCR - Parser Error'))
    dialog.setStyleSheet(get_style(self, 'dialog_window'))
    dialog.exec()



import gettext
from typing import Iterable


def init_translation(lang_code='en'):
    """
    Initialize translation.

    Parameters:
    - :param lang_code: Language codes, Example: 'en', 'zh', 'fr'

    :return: gettext translation function
    """
    if lang_code == 'en':
        return
    global translation_func
    translation_func = gettext.translation(
            'messages', localedir='locales', languages=[lang_code], fallback=True).gettext


def tr(message: str | Iterable[str]) -> str | tuple[str]:
    """
    Translates message into currently installed language. Accepts str or iterable of str. Iterables
    are returned as tuple object.
    """
    if isinstance(message, str):
        return translation_func(message)
    else:
        return tuple(translation_func(m) for m in message)
        # return map(translation_func, message)  # more memory efficient, but can't be indexed


def _identity(msg):
    return msg


if 'translation_func' not in globals():
    global translation_func
    translation_func = _identity
